//! # Tuxcord Reverse Proxy Header Authenthication
#![feature(
    anonymous_lifetime_in_impl_trait,
    iterator_try_collect,
    rwlock_downgrade,
    try_blocks
)]
#![allow(clippy::missing_errors_doc)]

use std::{
    fs::File,
    io::{self, Read, Write},
    net::{Shutdown, TcpListener, TcpStream},
    sync::Arc,
    thread,
};

use anyhow::Context;
use clap::Parser;
use thiserror::Error;

use crate::utils::headers::HeadersExt;

pub mod args;
pub mod config;
pub mod constants;
#[cfg(feature = "ipc")]
pub mod ipc;
pub mod utils;

use args::Args;

fn main() -> anyhow::Result<()> {
    let args = Args::parse();

    if !args.config.exists() {
        println!(
            "{:?} doesn't exist, creating a default config",
            &args.config
        );
        File::create(args.config)
            .context("failure creating the config file")?
            .write_all(constants::DEFAULT_CONFIG)
            .context("failure writting the contents to the config file")?;
        return Ok(());
    }

    let mut config_file = File::open(&args.config).context("failure opening the config file")?;
    let mut config = String::new();
    config_file
        .read_to_string(&mut config)
        .context("failure reading the config file")?;
    let config: config::Schema = toml::from_str(&config).context("invalid config file")?;

    #[cfg(feature = "ipc")]
    if let Some(args::Commands::Ipc) = args.command {
        ipc::start_client(config)?;
        return Ok(());
    }

    println!("config: {config:#?}");
    let listener = TcpListener::bind(config.listen_at).context("failure tcp listening")?;
    let config_arc = Arc::new(config); // will also serve as a counter

    #[cfg(feature = "ipc")]
    ipc::handle_daemon(config_arc.clone());

    for stream in listener.incoming() {
        match stream {
            Ok(mut client) => {
                let config_arc = config_arc.clone();
                thread::spawn(|| {
                    if let Err(err) = handle_client(&mut client, config_arc.as_ref()) {
                        eprintln!("err: invalid req head ({err}), closing...");
                        _ = client.shutdown(Shutdown::Both);
                    }
                    drop(client);
                    drop(config_arc);
                });
            }
            Err(err) => eprintln!("error with an incoming listener {err:#?}"),
        }
    }

    unreachable!("listener had to be killed unexpectedly");
}

#[derive(Error, Debug)]
pub enum ClientError {
    #[error("error reading request head: {0}")]
    HeadReadError(io::Error),
    #[error("failed to find \"host\" header")]
    HostHeaderNotFound,
    #[error("requested host {0:?}, but it's not registered")]
    HostNotRegistered(String),
    #[error("failed to connect to backend: {0}")]
    BackendConnectFail(io::Error),
    #[error("io error exchanging client and backend: {0}")]
    ExchangeIoError(io::Error),
}

fn handle_client(client: &mut TcpStream, config: &config::Schema) -> Result<(), ClientError> {
    use ClientError as E;

    let mut header_buf = [0u8; 1024 * 8];
    let mut read_pos = 0usize;

    let mut headers;
    let (pos, req) = loop {
        headers = [httparse::EMPTY_HEADER; 16];
        let mut req = httparse::Request::new(&mut headers);

        let n = client
            .read(&mut header_buf[read_pos..])
            .map_err(E::HeadReadError)?;
        read_pos += n;

        if let Ok(httparse::Status::Complete(n)) = req.parse(&header_buf[0..read_pos]) {
            break (n - b"\r\n\r\n".len(), req);
        }
    };

    let theres_body = req.headers.has_any(["conten-length", "transfer-encoding"]);

    let host_header = req.headers.get("host").ok_or(E::HostHeaderNotFound)?;

    // Now find that header and pas everything
    let read_hosts = config.hosts.read();
    let (addr, _, _) = read_hosts
        .get(host_header.as_ref())
        .ok_or_else(|| E::HostNotRegistered(host_header.to_string()))?;

    let mut stream = TcpStream::connect(addr).map_err(E::BackendConnectFail)?;
    drop(read_hosts);
    let r: io::Result<()> = try {
        stream.write_all(&header_buf[0..pos])?;
        // here we sent all original headers, append our own (TODO)

        // send our overhead
        stream.write_all(&header_buf[pos..read_pos])?;
        if theres_body {
            io::copy(client, &mut stream)?;
        }
        io::copy(&mut stream, client)?;
    };
    r.map_err(E::ExchangeIoError)?;

    Ok(())
}
